#!/usr/local/bin/ruby

require 'socket'
require 'timeout'

class Khttpsv_template
  #----------------------------------------------------------------------
  # class Khttpsv_template
  #   --Skeleton class for single thread HTTP/1.0 server
  #----------------------------------------------------------------------
  # INSTANCE VARIABLES:
  #
  # (HTTP/request)
  # @http_ver: Fixnum
  # @http_method: String
  # @uri_raw, @uri_encoded, @uri_query, @uri_encoded: String
  # @req_headers: Hash of Array of String
  # (HTTP/response)
  # @resp_headers: Hash of Array of String
  # @server_name: String
  # @server_addr: String
  # (Connection)
  # @sock: Socket
  # @loglevel: Fixnum
  #----------------------------------------------------------------------
  # METHODS:
  #
  #def dispatch_handler
  #def send_status_header(code, reason)
  #def send_response_header
  #def begin_content(body)
  #def send_content(body)
  #def initialize(sock)
  #def do_work
  #----------------------------------------------------------------------

  LOG_DEBUG = 0
  LOG_INFO = 1
  LOG_NOTICE = 2
  LOG_WARN = 3
  LOG_ERR = 4

  STAT_OK = 200
  STAT_MOVED_PERM = 301
  STAT_MOVED_TEMP = 302
  STAT_NOT_MODIFIED = 304
  STAT_BAD_REQUEST = 400
  STAT_UNAUTHORIZED = 401
  STAT_FORBIDDEN = 403
  STAT_NOT_FOUND = 404
  STAT_INTERNAL_ERR = 500
  STAT_NOT_IMPLEMENTED = 501
  STAT_SERVICE_UNAVAIL = 503

  #----------------------------------------------------------------------
  def Khttpsv_template.logger(s, lev = LOG_INFO)
    # class method version
    STDOUT.print "#{Thread.current.id}: #{s}\n"
  end
  def logger(s, lev = LOG_NOTICE)
    if lev >= @loglevel then
      Khttpsv_template.logger(s, lev)
    end
  end

  #----------------------------------------------------------------------
  RFC822_DAYS = %w[ Sun Mon Tue Wed Thu Fri Sat ]
  RFC822_MONTHS = %w[ Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec ]
  # thanks to cgi-lib.rb (^^)

  # generate RFC1123-style date-time string
  def Khttpsv_template.rfc1123date(time)
    # assuming "time" is instanse of Time
    time.strftime('%a, %d %b %Y %H:%M:%S GMT')
  end
  def rfc1123date(time) Khttpsv_template.rfc1123date(time) end

  # create Time object from RFC1123-style date-time string
  def Khttpsv_template.parse_rfc1123date(s)
    if s =~ %r{^\w+, (\d+) (\w+) (\d+) (\d+):(\d+):(\d+) GMT$}i then
      begin
	Time.gm($3.to_i, $2, $1.to_i, $4.to_i, $5.to_i, $6.to_i)
      rescue
	nil
      end
    else
      nil
    end
  end
  def parse_rfc1123date(s) Khttpsv_template.parse_rfc1123date(s) end

  #module_function :rfc1123date, :parse_rfc1123date

  #----------------------------------------------------------------------
  def Khttpsv_template.url_escape(s)
    s.gsub(/[^a-zA-Z0-9_\-.]/n){ sprintf("%%%02X", $1[0]) }
  end
  def url_escape(s) Khttpsv_template.url_escape(s) end

  def Khttpsv_template.url_unescape(s)
    s.gsub(/\+/, ' ').gsub(/%([0-9a-fA-F]{2})/) {$1.hex.chr}
  end
  def url_unescape(s) Khttpsv_template.url_unescape(s) end

  #module_function :url_escape, :url_unescape

  #----------------------------------------------------------------------
  def Khttpsv_template.html_escape(s)
    # escape plain text as HTML character entity refernece
    s.sub(/</, '&lt;').sub(/>/, '&gt;').sub(/&/, '&amp;')
  end
  def html_escape(s) Khttpsv_template.html_escape(s) end

  #module_function :html_escape

  #----------------------------------------------------------------------
  # parse request line
  def parse_request_line(req)
    if req =~ %r{ HTTP/(\d+)\.(\d+)$} then
      @http_ver = $1.to_i * 10 + $2.to_i # HTTP/1.0 or higher
      req = $`
    else
      @http_ver = 9
    end

    # split request line into method and URI
    unless tmp = req.index(' ') then
      logger("err malformed request line", LOG_ERR)
      return true
    end
    @http_method = req[0, tmp].upcase
    @uri_raw = req[tmp + 1, req.length - tmp]

    # split URI into path and query string
    if tmp = @uri_raw.index('?') then
      encoded = url_unescape(@uri_raw[0, tmp])
      @uri_query = @uri_raw[tmp + 1, @uri_raw.length - tmp]
    else
      encoded = url_unescape(@uri_raw)
      @uri_query = ''
    end

    # adjust '..' element in URI
    tmp = encoded.split('/', -1) # wont discard empty element
    for i in 1 .. tmp.length
      if tmp[i] == '..' then
	tmp[i - 1] = nil # deleted by compact
	tmp[i] = nil # deleted by compact
	tmp = tmp.compact
	redo
      end
    end
    @uri_encoded = tmp.compact.join('/')
    return false
  end

  #----------------------------------------------------------------------
  # read request header and store it into Hash (@req_headers)
  def recv_request_header
    return unless @http_ver >= 10

    # read request headers
    hdr = []
    while TRUE
      buf = ''
      begin
	status = timeout(120) do
	  buf = @sock.gets
	end
      end

      break if buf == ''
      buf.chop!
      break if buf == ''

      if buf =~ /^(\s+)/ then
	# continued header
	buf = buf[$1.length, -$1.length]
	hdr[-1] = hdr[-1] + ' ' + buf
      else
	hdr << buf
      end
    end # while

    # store them into @req_headers
    for buf in hdr
      next unless tmp = buf.index(':')
      key = buf[0, tmp].capitalize
      val = buf[tmp + 1, buf.length - tmp].sub(/^\s+/, '')
      if @req_headers[key].is_a?(Array) then
	@req_headers[key] << val
      else
	@req_headers[key] = [val]
      end
    end

    @req_headers.each do |key, val|
      logger('req_hdr ' + (key + ': ' + val.join("\n")).dump, LOG_DEBUG)
    end
  end

  #----------------------------------------------------------------------
  # 
  def parse_request_header
  end

  #----------------------------------------------------------------------
  #----------------------------------------------------------------------
  # 
  def send_status_header(code, reason)
    if @http_ver >= 10 then
      timeout(120) do
	@sock.print "HTTP/1.0 #{code} #{reason}\r\n"
      end
    end
    logger("response #{code} #{reason.dump}", LOG_INFO)
  end

  #----------------------------------------------------------------------
  # 
  def send_response_header
    if @http_ver >= 10 then
      # adds required fields
      @resp_headers['date'] = [rfc1123date(Time.now.gmtime)]
      @resp_headers['server'] = [@server_name]

      # send fields
      timeout(120) do
	@resp_headers.each do |key, val|
	  val.each do |val_item|
	    @sock.print "#{key.capitalize}: #{val_item}\r\n"
	  end
	end
      end

      @resp_headers.each do |key, val|
	logger('res_hdr ' + (key.capitalize + ': ' + val.join("\n")).dump, LOG_DEBUG)
      end
    end
  end

  #----------------------------------------------------------------------
  # 
  def begin_content
    send_content("\r\n")
  end

  #----------------------------------------------------------------------
  # 
  def send_content(body)
    # disconnect if band-width < 10bytes/sec
    timeout = body.length / 10
    timeout = 120 if(timeout < 120)
    timeout(timeout) do
      @sock.print body
    end
  end

  #----------------------------------------------------------------------
  # 
  def dispatch_handler
    send_status_header(STAT_NOT_IMPLEMENTED, 'This is template server')
  end

  #----------------------------------------------------------------------
  #----------------------------------------------------------------------
  # 
  def do_work
    logger(sprintf('client %s/%s', @sock.peeraddr[2], @sock.peeraddr[3]), LOG_NOTICE)

    begin
      req = '*invalid request*'
      timeout(120) do
	req = @sock.gets
	req.chop!
      end
      logger('request ' + req.dump, LOG_NOTICE)

      parse_request_line(req)
      recv_request_header
      parse_request_header

      dispatch_handler

    rescue TimeoutError
      logger('err timed out', LOG_ERR)

    rescue IOError
      logger('err IO error', LOG_ERR)

    rescue Errno::EPIPE
      logger('err broken pipe', LOG_ERR)
    end
  end

  #----------------------------------------------------------------------
  # 
  def initialize(s)
    @sock = s
    @loglevel = LOG_INFO
    @server_name = 'TemplateRubyHTTPServer/1.0.0'
    @req_headers = {}
    @resp_headers = {}
  end
end
